#!/usr/bin/env python
#
#   zfilenet/blobserver.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""
Blob server implementation using SQLite3.

Limits on SQLite3:

* Maximum column size: 1,000,000 bytes (approximately 953 MiB)
* Maximum database size: 2^15 * (2^30 - 1) (approximately 31TiB)
* Recommended database size: <24GiB
* Memory footprint: 256 bytes for every 1 MiB of data (depending on page size)
"""

from __future__ import absolute_import
from base64 import b64decode, b64encode
import logging
import sys

import pinet
from pinet.dispatch import expose
import progtools
from sqlalchemy import (Table, Column, MetaData,
                        Integer, String, Binary, ForeignKey)
from sqlalchemy.orm import mapper, sessionmaker, synonym
from twisted.internet import reactor

from . import db, discovery, util

__author__ = 'Ross Light'
__date__ = 'June 13, 2008'
__version__ = '0.1.3'
__all__ = ['usage',
           'description',
           'log',
           'create_session',
           'DEFAULT_DB_FILE',
           'MAX_BLOB_SIZE',
           'DEFAULT_QUOTA',
           'DataBlob',
           'BlobServer',
           'connect',
           'main',]

usage = "usage: %prog [-p PORT] [-n NAME] [database]"
description = "Storage server for ZFN"
log = logging.getLogger('zfilenet.blobserver')
create_session = sessionmaker(autoflush=True, transactional=True)

DEFAULT_DB_FILE = 'zfn_blob.db'
MAX_BLOB_SIZE = 128 * (1024)    # 128 KiB
DEFAULT_QUOTA = 2 * (1024 ** 3) # 2 GiB

### TABLES ###

metadata = MetaData()
data_blobs_table = Table('data_blobs', metadata,
    Column('blob_db_id', Integer, primary_key=True),
    Column('blob_uuid', String(32), unique=True),
    Column('blob_data', Binary(MAX_BLOB_SIZE)),
    Column('blob_size', Integer),
)

### MAPPING CLASSES ###

class DataBlob(object):
    """
    A single "blob" of data, usually a segment of a file, identified by an ID.
    
    :IVariables:
        uuid : uuid.UUID
            The blob's unique identifier
        data : str
            The binary data the blob contains
        size : int
            (Read-Only) The length of the data (in bytes)
    """
    @classmethod
    def fetch_blob(cls, session, blob_id):
        """
        Fetches the blob identified with an ID.
        
        :Parameters:
            session : sqlalchemy.orm.session.Session
                The session to make queries from
            blob_id : UUID-like
                The ID of the blob to fetch
        :Returns: The blob with the given ID, or ``None`` if not found
        :ReturnType: DataBlob
        """
        blob_id = util.to_uuid(blob_id).hex
        return session.query(cls).filter_by(uuid=blob_id).first()
    
    @classmethod
    def read_blob(cls, session, blob_id):
        """
        Reads the data from the blob identified with an ID.
        
        :Parameters:
            session : sqlalchemy.orm.session.Session
                The session to make queries from
            blob_id : UUID-like
                The ID of the blob to read from
        :Returns: The requested data, or ``None`` if not found
        :ReturnType: str
        """
        blob = cls.fetch_blob(session, blob_id)
        return blob.data if blob is not None else None
    
    @classmethod
    def write_blob(cls, session, blob_id, data):
        """
        Writes data to a new blob.
        
        :Parameters:
            session : sqlalchemy.orm.session.Session
                The session to make queries from
            blob_id : UUID-like
                The ID of the blob to create
            data : str
                The data to write to the blob
        :Raises KeyError: If a blob with the given ID already exists
        :Returns: The newly created blob
        :ReturnType: DataBlob
        """
        blob = cls.fetch_blob(session, blob_id)
        if blob is not None:
            raise KeyError(blob_id)
        else:
            blob = cls(blob_id, data)
            session.save(blob)
        session.commit()
        return blob
    
    @classmethod
    def delete_blob(cls, session, blob_id):
        """
        Deletes the blob identified with an ID.
        
        :Parameters:
            session : sqlalchemy.orm.session.Session
                The session to make queries from
            blob_id : UUID-like
                The ID of the blob to delete
        :Raises KeyError: If a blob does not exist with the ID
        """
        blob = cls.fetch_blob(session, blob_id)
        if blob is not None:
            session.delete(blob)
            session.commit()
        else:
            raise KeyError(blob_id)
    
    def __init__(self, blob_id, data):
        self.uuid = blob_id
        self.data = data
    
    def __repr__(self):
        return "<DataBlob urn:blob:%s>" % (self.uuid)
    
    def __str__(self):
        return "urn:blob:%s" % self.uuid
    
    # UUID property
    def _get_uuid(self):
        return util.to_uuid(self._uuid)
    def _set_uuid(self, value):
        self._uuid = util.to_uuid(value).hex
    uuid = property(_get_uuid, _set_uuid)
    
    # Data property
    def _get_data(self):
        return str(self._data)
    def _set_data(self, value):
        self._data = buffer(value)
        self._size = len(value)
    data = property(_get_data, _set_data)
    
    @property
    def size(self):
        return self._size

### MAPPERS ###

mapper(DataBlob, data_blobs_table, properties=dict(
    _uuid=data_blobs_table.c.blob_uuid,
    uuid=synonym('_uuid'),
    _data=data_blobs_table.c.blob_data,
    data=synonym('_data'),
    _size=data_blobs_table.c.blob_size,
    size=synonym('_size'),
))

### SERVER LOGIC ###

def _get_space_used(session):
    """
    Calculates the amount of space used in the database.
    
    Notice that this is almost certainly not the size of the database on disk;
    it is merely the sum of all the blob sizes.
    
    :Parameters:
        session : sqlalchemy.orm.session.Session
            The session to make queries from
    :Returns: The number of bytes used in the database
    :ReturnType: int
    """
    size_sum = session.query(DataBlob).sum(data_blobs_table.c.blob_size)
    return size_sum if size_sum is not None else 0

class BlobServer(object):
    """
    The JSON-RPC interface to the server.
    
    :IVariables:
        engine : sqlalchemy.engine.Engine
            The database connection
        quota : int
            The amount of space the server is allowed to use (in bytes)
    """
    def __init__(self, engine, quota):
        self.engine = engine
        self.quota = quota
    
    def _get_session(self):
        return create_session(bind=self.engine)
    
    @expose()
    def get_version(self):
        """
        Retrieves the version of the server.
        
        :Returns: The server version string
        :ReturnType: str
        """
        return __version__
    
    @expose()
    def read(self, blob_id):
        """
        Reads from a blob.
        
        :Parameters:
            blob_id : str
                The blob's ID as a string
        :Returns: Base64-encoded data from the blob, or ``None`` if not found
        :ReturnType: str
        """
        session = self._get_session()
        blob = DataBlob.read_blob(session, blob_id)
        return b64encode(blob) if blob is not None else blob
    
    @expose()
    def write(self, blob_id, data):
        """
        Writes to a blob.
        
        ==========  ===================================================
        Error Code  Reason
        ==========  ===================================================
        -1          Blob already exists
        -2          Blob is larger than the limit
        -3          Adding this blob would put the server over quota
        ==========  ===================================================
        
        :Parameters:
            blob_id : str
                The blob's ID as a string
            data : str
                Base64-encoded data to write to the blob
        :Raises pinet.jsonrpc.RPCError: If there was an error
        """
        session = self._get_session()
        data = b64decode(data)
        current_size = _get_space_used(session)
        # Check for over-sizeness
        if len(data) > MAX_BLOB_SIZE:
            raise pinet.jsonrpc.RPCError(-2, "Blob is too large")
        elif current_size + len(data) > self.quota:
            log.warning("Server is over quota")
            raise pinet.jsonrpc.RPCError(-3, "Over quota")
        # Write the blob
        try:
            DataBlob.write_blob(session, blob_id, data)
            log.debug("Wrote %i bytes to %s", len(data), blob_id)
        except KeyError:
            raise pinet.jsonrpc.RPCError(-1, "Blob already exists")
    
    @expose()
    def get_blob_list(self, query=None):
        """
        Fetches the list of blobs present on the server.
        
        :Parameters:
            query : list of str
                (Optional) The limiter for the list.  If given, the server will
                only return what subset of that list is present.
        :Returns: Which blob IDs are present
        :ReturnType: list of str
        """
        session = self._get_session()
        results = session.query(DataBlob)
        if query is not None:
            query = [util.to_uuid(blob_id).hex for blob_id in query]
            results = results.filter(data_blobs_table.c.blob_uuid.in_(query))
        return [str(blob.uuid) for blob in results]
    
    @expose()
    def get_limits(self):
        """
        Retrieves the limits of the server.
        
        This returns a tuple that contains (in this order): the maximum size
        for a single blob, the space used, and the maximum quota.
        
        :Returns: The server's limits
        :ReturnType: tuple
        """
        session = self._get_session()
        return (MAX_BLOB_SIZE, _get_space_used(session), self.quota)

def _on_new(engine):
    log.debug("Creating new database...")
    session = create_session(bind=engine)
    db.DBVariable.set_variable(session, 'version', __version__)

def connect(db_file=DEFAULT_DB_FILE):
    """
    Connects to a blob database file.
    
    :Parameters:
        db_file : str
            The path to the database file.  If no path is given, then the
            default name is used.
    :Returns: The database engine
    :ReturnType: sqlalchemy.engine.Engine
    """
    return db.connect_sqlite3(metadata, db_file, _on_new)

@progtools.reportexits
def main(args=None):
    """
    Runs the blob server.
    
    :Parameters:
        args : list of str
            Command-line arguments.  If not given, ``sys.argv[1:]`` is used.
    :Returns: The exit code of the program
    :ReturnType: int
    """
    # Parse arguments
    if args is None:
        args = sys.argv[1:]
    parser = progtools.OptionParser(usage=usage, version=__version__,
                                    description=description,)
    parser.set_defaults(port=8000,
                        service_name=None,
                        quota=DEFAULT_QUOTA,)
    util.add_log_options(parser)
    parser.add_option('-p', '--port',
                      action='store',
                      dest='port',
                      type='int',
                      metavar="PORT",
                      help="Changes the port the server runs on. "
                           "[default: %default]")
    parser.add_option('-n', '--service-name',
                      action='store',
                      dest='service_name',
                      metavar="NAME",
                      help="The service name broadcast with Bonjour. "
                           "[default: host name]")
    parser.add_option('--quota',
                      action='store',
                      dest='quota',
                      type='int',
                      metavar="BYTES",
                      help="The maximum amount of storage. "
                           "[default: 2GiB]")
    options, args = parser.parse_args(args)
    util.setup_log_options(options)
    if len(args) == 0:
        db_file = DEFAULT_DB_FILE
    elif len(args) == 1:
        db_file = args[0]
    else:
        raise UsageError()
    # Set up server
    log.debug("Using database: %s", db_file)
    engine = connect(db_file)
    peer = BlobServer(engine, options.quota)
    factory = pinet.jsonrpc.JSON_RPCFactory()
    factory.requests.register_object(peer)
    # Show some statistics
    session = create_session(bind=engine)
    log.debug("Database has %i blobs", session.query(DataBlob).count())
    log.debug("Database using %s",
              util.format_file_size(_get_space_used(session)))
    del session
    # Set up Bonjour broadcasting
    log.info("Establishing Bonjour services...")
    discovery.broadcast(reactor, discovery.BLOB_SERVICE_TYPE,
                        options.port, options.service_name)
    # Run reactor
    log.debug("Acquiring TCP port %i...", options.port)
    reactor.listenTCP(options.port, factory)
    log.info("Running...")
    reactor.run()
    return 0

if __name__ == '__main__':
    sys.exit(main())
