#!/usr/bin/env python
#
#   zfilenet/indexserver/dav.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""WebDAV access to a ZFN share."""

from __future__ import absolute_import
from base64 import b64encode, b64decode
from cgi import escape
from hashlib import sha1
import logging
import mimetypes
import posixpath
import re
from tempfile import TemporaryFile
from urllib import quote, quote_plus, unquote
from urlparse import urlparse
from uuid import UUID, uuid4
from xml.etree import ElementTree

import genshi
from genshi.template import TemplateLoader
import pkg_resources
from sqlalchemy.orm import object_session
from twisted.web import error, http, resource, server
from twisted.web.static import addSlash, FileTransfer
from twisted.web.util import redirectTo

from . import index
from .. import discovery, util

__author__ = 'Ross Light'
__date__ = 'June 27, 2008'
__all__ = ['DAVResponse',
           'DAVProperty',
           'DirectoryListing',
           'DirectoryCreator',
           'FilePutter',
           'StaticNamespace',
           'ZFNResource',
           'create_server',]

log = logging.getLogger('zfilenet.indexserver.dav')
_template_dir = pkg_resources.resource_filename('zfilenet', 'templates')
_template_loader = TemplateLoader([_template_dir])
DAV_NAMESPACE = 'DAV:'

def _render_template(template, values=None, **kw):
    method = kw.pop('genshi_format', 'html')
    if values is None:
        values = {}
    else:
        values = dict(values)
    values.update(kw)
    zfn_ns = dict(quote=quote,
                  format_file_size=util.format_file_size,
                  format_datetime=util.format_datetime,
                  static=_render_static,)
    values['zfn'] = zfn_ns
    template = _template_loader.load(template)
    result_stream = template.generate(**values)
    return result_stream.render(method)

def _render_error(code, message=None):
    return _render_template("error.html",
                            code=code,
                            meaning=http.RESPONSES[code],
                            message=message,)

def _render_static(name, mime_type=None):
    return '/zfn-static?name=%s' % (quote_plus(name))

def _send_error(request, code, message=None):
    request.setResponseCode(code)
    request.setHeader('Content-Type', 'text/html')
    return _render_error(code, message)

def _get_host(request):
    port = request.getHost().port
    return 'http://' + request.getRequestHostname() + ':' + str(port)

def _get_mime_type(index_obj):
    mime_type, encoding = mimetypes.guess_type(index_obj.name)
    return mime_type

def _format_http_date(dt):
    return dt.strftime("%a, %d %b %Y %H:%M:%S GMT")

def _get_properties(properties, index_obj):
    # Generate all properties
    item_properties = [
        DAVProperty(DAV_NAMESPACE, 'creationdate',
                    util.format_datetime(index_obj.created)),
        DAVProperty(DAV_NAMESPACE, 'getlastmodified',
                    _format_http_date(index_obj.modified)),]
    if index_obj.is_file:
        item_properties += [
            DAVProperty(DAV_NAMESPACE, 'getcontentlength', index_obj.length),
            DAVProperty(DAV_NAMESPACE, 'getcontenttype', _get_mime_type(index_obj)),
            DAVProperty(DAV_NAMESPACE, 'resourcetype', None),
        ]
    elif index_obj.is_directory:
        item_properties += [
            DAVProperty(DAV_NAMESPACE, 'resourcetype',
                        '<D:collection/>', escape_value=False),
        ]
    # Return correct properties
    if properties in ('all', 'names'):
        return item_properties, [], []
    else:
        found, not_found, forbidden = [], [], []
        for prop in properties:
            for item_prop in item_properties:
                if item_prop.namespace == prop.namespace and \
                   item_prop.name == prop.name:
                    found.append(item_prop)
                    break
            else:
                not_found.append(prop)
        return (found, not_found, forbidden)

def _set_properties(properties, index_obj):
    successful, forbidden, conflict = [], [], []
    for prop_list in properties:
        optype, prop_list = prop_list[0], prop_list[1:]
        for prop in prop_list:
            if prop.name in ('getcontentlength', 'getcontenttype',
                             'getlastmodified', 'resourcetype',
                             'creationdate'):
                # Automatically generated properties
                conflict.append(prop)
            else:
                # We don't like other properties
                forbidden.append(prop)
    # Return results
    if forbidden or conflict:
        return [], forbidden, conflict, successful
    else:
        return successful, [], [], []

def _write_file(session, blob_server_list, path, content):
    BUFFER_SIZE = 8 * 1024 # 8 KiB
    BLOB_SIZE = 128 * 1024 # 128 KiB
    # Determine mode
    node = index.resolve_path(session, path)
    append = False
    if node is not None:
        # File does exist, check for appending
        if node.is_directory:
            # Trying to write to a directory?  Shame.
            return False
        file_hash = sha1()
        hash_length = 0
        while content.tell() < node.length:
            data = content.read(BUFFER_SIZE)
            if not data:
                break
            update_length = min(node.length - hash_length, BUFFER_SIZE)
            file_hash.update(data[:update_length])
            hash_length += len(data)
        else:
            if file_hash.digest() == node.sha1_hash:
                append = True
        if append:
            content.seek(node.length)
            file_length = node.length
        else:
            content.seek(0)
            node.clear()
            file_hash = sha1()
            file_length = 0
    else:
        # File does not exist, create it.
        try:
            node = index.File.make(session, path)
        except ValueError:
            return False
        else:
            file_hash = sha1()
            file_length = 0
    # Write data
    writer = blob_server_list.get_writer()
    while True:
        data = content.read(BLOB_SIZE)
        if not data:
            break
        new_blob_id = uuid4()
        node.add_blob(new_blob_id)
        file_hash.update(data)
        file_length += len(data)
        writer.write_blob(new_blob_id, data)
    # Finish it!
    node.length = file_length
    node.sha1_hash = file_hash.digest()
    session.commit()
    return True

class DAVResponse(object):
    def __init__(self, href, properties=None, not_found=None, forbidden=None,
                 conflict=None, failed=None):
        self.href = href
        self.properties = properties if properties is not None else []
        self.not_found = not_found if not_found is not None else []
        self.forbidden = forbidden if forbidden is not None else []
        self.conflict = conflict if conflict is not None else []
        self.failed = failed if failed is not None else []

class DAVProperty(object):
    _name_pattern = re.compile(r'^[_:a-zA-Z][-_:.a-zA-Z0-9]*$')
    
    def __init__(self, *args, **kw):
        # Get arguments
        if len(args) == 2:
            (name, value), namespace = args, None
        elif len(args) == 3:
            namespace, name, value = args
        else:
            raise TypeError("DAVProperty.__init__ takes 2-3 arguments")
        escape_value = kw.pop('escape_value', True)
        if kw:
            raise TypeError("Unrecognized keyword parameter given to "
                            "DAVProperty.__init__")
        # Initialize
        if not self._name_pattern.match(name):
            raise ValueError("Bad property name")
        self.name = name
        self.value = unicode(value) if value is not None else None
        self.namespace = namespace
        self.escape_value = bool(escape_value and self.value is not None)
    
    def __repr__(self):
        if self.namespace is None:
            return "DAVProperty(%r, %r)" % (self.name, self.value)
        else:
            return "DAVProperty(%r, %r, %r)" % \
                (self.namespace, self.name, self.value)
    
    def __str__(self):
        return unicode(self).encode()
    
    def __unicode__(self):
        return self.render()
    
    def render(self, name_only=False):
        """
        Render the property to a string.
        
        :Keywords:
            name_only : bool
                Whether to only generate the name
        :Returns: The XML data of the property
        :ReturnType: unicode
        """
        value = escape(self.value) if self.escape_value else self.value
        values = {'name': self.name,
                  'value': value,}
        if self.namespace is not None:
            values['ns'] = escape(self.namespace, quote=True)
            values['prefix'] = 'D' if self.namespace == DAV_NAMESPACE else 'N'
            format = u"<%(prefix)s:%(name)s xmlns:%(prefix)s=\"%(ns)s\""
        else:
            format = u"<%(name)s"
        if name_only or value is None:
            format += u" />"
        else:
            format += u">%(value)s</"
            if 'prefix' in values:
                format += "%(prefix)s:"
            format += "%(name)s>"
        return format % values
    
    def serialize(self, name_only=False):
        """
        Render the property to a string.
        
        :Keywords:
            name_only : bool
                Whether to only generate the name
        :Returns: The XML data of the property
        :ReturnType: unicode
        """
        stream = genshi.XML(self.render(name_only=name_only))
        # We run this through a namespace flattener to make sure we get the
        # correct namespacing.
        return stream | genshi.output.NamespaceFlattener()

class DirectoryListing(resource.Resource):
    def __init__(self, directory):
        resource.Resource.__init__(self)
        self.directory = directory
        self.isLeaf = True
    
    def render_GET(self, request):
        session = object_session(self.directory)
        request.setHeader('Content-Type', 'text/html')
        request.setHeader('Last-Modified',
                          _format_http_date(self.directory.modified))
        is_root = bool(self.directory == index.get_root(session))
        return _render_template("listing.html",
                                request=request,
                                directory=self.directory,
                                is_root=is_root,)

class DirectoryCreator(resource.Resource):
    def __init__(self, directory, name, bad_path=False):
        resource.Resource.__init__(self)
        self.isLeaf = False
        self.directory = directory
        self.name = name
        self.bad = bad_path
    
    def getChild(self, path, request):
        if path:
            return DirectoryCreator(self.directory, path, True)
        else:
            return self
    
    def render_MKCOL(self, request):
        session = object_session(self.directory)
        if self.bad:
            return _send_error(request, http.CONFLICT)
        elif self.directory.get_child(self.name) is not None:
            return _send_error(request, http.NOT_ALLOWED)
        else:
            path = self.directory.get_path() + self.name
            log.debug("Creating directory %s from MKCOL", path)
            index.Directory.make(session, path)
            request.setResponseCode(http.CREATED)
            return ''

class FilePutter(resource.Resource):
    def __init__(self, blob_server_list, directory, name, bad_path=False):
        resource.Resource.__init__(self)
        self.isLeaf = False
        self.blob_server_list = blob_server_list
        self.directory = directory
        self.name = name
        self.bad = bad_path
    
    def getChild(self, path, request):
        return FilePutter(blob_server_list, self.directory, path, True)
    
    def render_PUT(self, request):
        session = object_session(self.directory)
        if self.bad:
            return _send_error(request, http.CONFLICT)
        elif self.directory.get_child(self.name) is not None:
            return _send_error(request, http.NOT_ALLOWED)
        else:
            path = self.directory.get_path() + self.name
            log.debug("Creating file %s from PUT", path)
            success = _write_file(session, self.blob_server_list,
                                  path, request.content)
            if success:
                request.setResponseCode(http.CREATED)
                return ''
            else:
                return _send_error(request, http.CONFLICT)

class StaticNamespace(resource.Resource):
    resources = {'arrow': 'static/arrow.png',
                 'header': 'static/header.png',
                 'logo-16': 'static/logo-16.png',
                 'logo-64': 'static/logo-64.png',
                 'parent': 'static/parent.png',
                 'common-stylesheet': 'static/common.css',
                 'zfn-stylesheet': 'static/zfn.css',
                 'zfn-stylesheet-iphone': 'static/zfn_iphone.css',}
    isLeaf = True
    
    def render_GET(self, request):
        resource_name = request.args['name'][0]
        try:
            resource_location = self.resources[resource_name]
        except KeyError:
            return _send_error(request, http.NOT_FOUND)
        else:
            fileobj = pkg_resources.resource_stream('zfilenet',
                                                    resource_location)
            # Get file size
            fileobj.seek(0, 2)
            size = fileobj.tell()
            fileobj.seek(0)
            request.setHeader('Content-Length', str(size))
            # Get file type
            mime_type, encoding = mimetypes.guess_type(resource_location)
            if mime_type:
                request.setHeader('Content-Type', mime_type)
            if encoding:
                request.setHeader('Content-Encoding', encoding)
            # Handle HEAD
            if request.method == 'HEAD':
                return ''
            # Handle GET
            FileTransfer(fileobj, size, request)
            return server.NOT_DONE_YET

class ZFNResource(resource.Resource):
    _child_not_found = error.NoResource("File not found.")
    
    def __init__(self, engine, blob_server_list, index_obj=None):
        resource.Resource.__init__(self)
        self.engine = engine
        self.blob_server_list = blob_server_list
        self.index_obj = index_obj
        self.isLeaf = bool(self.index_obj is not None and
                           not self.index_obj.is_directory)
    
    def _get_session(self):
        # Determine if we have an object session
        if self.index_obj is not None:
            session = object_session(self.index_obj)
        else:
            session = None
        # If we have no session, create one
        if session is None:
            session = index.create_session(bind=self.engine)
        return session
    
    def _get_index_obj(self, session):
        if self.index_obj is None:
            return index.get_root(session)
        else:
            return self.index_obj
    
    def _get_data(self):
        session = self._get_session()
        index_obj = self._get_index_obj(session)
        return session, index_obj
    
    def listDynamicNames(self):
        session, index_obj = self._get_data()
        if index_obj.is_directory:
            return [''] + [c.name for c in index_obj.get_children()]
        else:
            return []
    
    def getChild(self, path, request):
        session, index_obj = self._get_data()
        # Special handling for /zfn-resource
        if path == 'zfn-static' and \
           self.index_obj is None and \
           request.args:
            return StaticNamespace()
        # Now for normal children
        if index_obj.is_directory:
            if not path:
                if request.method in ('HEAD', 'GET'):
                    return DirectoryListing(index_obj)
                else:
                    return self
            else:
                child = index_obj.get_child(path)
                if child is not None:
                    return ZFNResource(self.engine, self.blob_server_list,
                                       child)
                elif request.method == 'MKCOL':
                    return DirectoryCreator(index_obj, path)
                elif request.method == 'PUT':
                    return FilePutter(self.blob_server_list, index_obj, path)
                else:
                    return self._child_not_found
        else:
            return self._child_not_found
    
    def render(self, request):
        if request.method not in ('HEAD', 'GET', 'PROPFIND'):
            log.debug("ZFN: Client attempting %s on %s",
                      request.method, request.path)
        return resource.Resource.render(self, request)
    
    def render_OPTIONS(self, request):
        request.setHeader('DAV', '1')
        return ''
    
    def render_GET(self, request):
        session, index_obj = self._get_data()
        log.debug("GETing %s", request.path)
        if index_obj.is_directory:
            # Directory without trailing slash
            return self.index_redirect(request)
        if index_obj.length == 0 or not index_obj.blob_list:
            # Empty file
            request.setHeader('Content-Type', _get_mime_type(index_obj))
            request.setHeader('Last-Modified',
                              _format_http_date(index_obj.modified))
            return ''
        # Serving regular file
        log.debug("Serving file %s (UUID %s)",
                  index_obj.name, index_obj.uuid)
        fetcher = self.blob_server_list.fetch_list(index_obj.blob_list)
        fetcher.has_blobs.addCallback(self._blobs_mapped,
                                      request, session, index_obj)
        fetcher.progress.addCallback(self._blob_fetched,
                                     request, session, index_obj)
        fetcher.finished.addCallback(self._blobs_done,
                                     request, session, index_obj)
        fetcher.start()
        return server.NOT_DONE_YET
    
    def index_redirect(self, request):
        return redirectTo(addSlash(request), request)
    
    def _blobs_mapped(self, success, request, session, index_obj):
        log.debug("Finished mapping (success: %i)", success)
        if success:
            # Write headers
            request.setHeader('Content-Type', _get_mime_type(index_obj))
            request.setHeader('Last-Modified',
                              _format_http_date(index_obj.modified))
            request.setHeader('Content-Length', index_obj.length)
        else:
            # Not all blobs found
            code = http.SERVICE_UNAVAILABLE
            request.setResponseCode(code)
            data = _render_error(code, "Not all blobs present")
            request.setHeader('Content-Type', 'text/html')
            request.setHeader('Content-Length', len(data))
            request.write(data)
    
    def _blob_fetched(self, result, request, session, index_obj):
        blob_id, blob = result
        request.write(blob)
    
    def _blobs_done(self, success, request, session, index_obj):
        log.debug("Finished fetching (success: %i)", success)
        if not success:
            code = http.SERVICE_UNAVAILABLE
            request.setResponseCode(code)
            data = _render_error(code, "Not all blobs present")
            request.setHeader('Content-Type', 'text/html')
            request.setHeader('Content-Length', len(data))
            request.write(data)
        request.finish()
    
    def render_PROPFIND(self, request):
        session, index_obj = self._get_data()
        log.debug("PROPFINDing %s", request.path)
        request_info = self._parse_propfind(request)
        responses = self._find_properties(_get_host(request),
                                          unquote(request.path),
                                          index_obj,
                                          request_info['properties'],
                                          request_info['depth'])
        request.setResponseCode(http.MULTI_STATUS)
        request.setHeader('Content-Type', 'text/xml')
        name_only = bool(request_info['properties'] == 'names')
        return _render_template("dav_multistatus.xml",
                                genshi_format='xml',
                                responses=responses,
                                name_only=name_only,)
    
    def _parse_propfind(self, request):
        # Start result dictionary
        result = {}
        result['depth'] = request.getHeader('Depth')
        if result['depth'] is None:
            result['depth'] = 'infinity'
        # Parse request
        if request.getHeader('Content-Length') == '0':
            result['properties'] = 'all'
        else:
            doc = ElementTree.parse(request.content)
            root = doc.getroot()
            # TODO: Check if root is {DAV:}propfind
            prop_root = root.find('{DAV:}prop')
            # Find properties
            # TODO: Raise error if allprop and propname are both present
            if prop_root.find('{DAV:}allprop') is not None:
                result['properties'] = 'all'
            elif prop_root.find('{DAV:}propname') is not None:
                result['properties'] = 'names'
            else:
                result['properties'] = self._parse_props(prop_root)
        return result
    
    @staticmethod
    def _parse_props(elem):
        result = []
        for prop_elem in elem:
            namespace, name = util.parse_qname(prop_elem.tag)
            new_property = DAVProperty(namespace, name, None)
            result.append(new_property)
        return result
    
    @classmethod
    def _find_properties(cls, host, path, index_obj, properties, depth='0'):
        responses = []
        found, not_found, forbidden = _get_properties(properties, index_obj)
        responses.append(DAVResponse(host + quote(path),
                                     found, not_found, forbidden))
        if index_obj.is_file or depth == '0':
            pass
        else:
            for child in index_obj.get_children():
                child_depth = '0' if depth == '1' else 'infinity'
                result = cls._find_properties(host,
                                              posixpath.join(path, child.name),
                                              child, properties, child_depth)
                responses += result
        return responses
    
    def render_PROPPATCH(self, request):
        session, index_obj = self._get_data()
        request_info = self._parse_proppatch(request)
        properties, forbidden, conflict, failed = \
            _set_properties(request_info, index_obj)
        response = DAVResponse(_get_host(request) + quote(request.path),
                               properties=properties,
                               forbidden=forbidden,
                               conflict=conflict,
                               failed=failed,)
        request.setResponseCode(http.MULTI_STATUS)
        request.setHeader('Content-Type', 'text/xml')
        return _render_template("dav_multistatus.xml",
                                genshi_format='xml',
                                responses=[response],
                                name_only=True,)
    
    def _parse_proppatch(self, request):
        result = []
        # Parse request
        doc = ElementTree.parse(request.content)
        root = doc.getroot()
        # TODO: Check if root is {DAV:}propertyupdate
        for child in root:
            if child.tag == '{DAV:}set':
                prop_elem = child.find('{DAV:}prop')
                result.append(['set'] + self._parse_props(prop_elem))
            elif child.tag == '{DAV:}remove':
                prop_elem = child.find('{DAV:}remove')
                result.append(['remove'] + self._parse_props(prop_elem))
        return result
    
    def render_MKCOL(self, request):
        return _send_error(request, http.NOT_ALLOWED)
    
    def render_DELETE(self, request):
        session, index_obj = self._get_data()
        if index_obj == index.get_root(session):
            return _send_error(request, http.FORBIDDEN,
                               "Root cannot be deleted")
        else:
            index_obj.delete(force=True) 
            code = http.NO_CONTENT
            request.setResponseCode(code)
            return ''
    
    def render_PUT(self, request):
        session, index_obj = self._get_data()
        if index_obj.is_directory:
            return _send_error(request, http.CONFLICT,
                               "Directories cannot be put")
        else:
            path = index_obj.get_path()
            log.debug("Modifying file %s from PUT", path)
            success = _write_file(session, self.blob_server_list, path,
                                  request.content)
            if success:
                request.setResponseCode(http.NO_CONTENT)
                return ''
            else:
                return _send_error(request, http.CONFLICT)
    
    def render_COPY(self, request):
        session, index_obj = self._get_data()
        request_info = self._parse_copy(request)
        host = _get_host(request)
        dest_url = urlparse(request_info['destination'])
        dest_host = dest_url[0] + '://' + dest_url[1]
        dest_path = unquote(dest_url[2])
        # Check for foreign server
        if host != dest_host:
            return _send_error(request, http.BAD_GATEWAY,
                               "This server does not support cross-server "
                               "copying.")
        # Check for root copying
        if index_obj == index.get_root(session):
            return _send_error(request, http.FORBIDDEN,
                               "Root cannot be copied")
        # Check overwriting
        dest_node = index.resolve_path(session, dest_path)
        overwrote = False
        if dest_node is not None:
            if request_info['overwrite']:
                dest_node.delete(force=True)
                overwrote = True
            else:
                return _send_error(request, http.PRECONDITION_FAILED,
                                   "Destination exists")
        # Copy
        try:
            index.move(session, index_obj.get_path(), dest_path,
                       copy=True,
                       recursive=bool(request_info['depth'] == 'infinity'),
                       allow_directory=False)
        except:
            return _send_error(request, http.FORBIDDEN)
        else:
            if overwrote:
                request.setResponseCode(http.NO_CONTENT)
            else:
                request.setResponseCode(http.CREATED)
            return ''
    
    def render_MOVE(self, request):
        session, index_obj = self._get_data()
        request_info = self._parse_copy(request)
        host = _get_host(request)
        dest_url = urlparse(request_info['destination'])
        dest_host = dest_url[0] + '://' + dest_url[1]
        dest_path = unquote(dest_url[2])
        # Check for foreign server
        if host != dest_host:
            return _send_error(request, http.BAD_GATEWAY,
                               "This server does not support cross-server "
                               "copying.")
        # Check for root copying
        if index_obj == index.get_root(session):
            return _send_error(request, http.FORBIDDEN,
                               "Root cannot be moved")
        # Check for bad depth
        if request_info['depth'] != 'infinity':
            return _send_error(request, http.BAD_REQUEST,
                               "Depth other than infinity given")
        # Check overwriting
        dest_node = index.resolve_path(session, dest_path)
        overwrote = False
        if dest_node is not None:
            if request_info['overwrite']:
                dest_node.delete(force=True)
                overwrote = True
            else:
                return _send_error(request, http.PRECONDITION_FAILED,
                                   "Destination exists")
        # Move
        try:
            index.move(session, index_obj.get_path(), dest_path,
                       allow_directory=False)
        except:
            return _send_error(request, http.FORBIDDEN)
        else:
            if overwrote:
                request.setResponseCode(http.NO_CONTENT)
            else:
                request.setResponseCode(http.CREATED)
            return ''
    
    def _parse_copy(self, request):
        # Start result dictionary
        result = {}
        result.setdefault('omit', False)
        result.setdefault('keepalive', None)
        # Get headers
        result['depth'] = request.getHeader('Depth')
        if result['depth'] is None:
            result['depth'] = 'infinity'
        overwrite = request.getHeader('Overwrite')
        overwrite_values = {'T': True, 'F': False}
        result['overwrite'] = overwrite_values.get(overwrite, False)
        result['destination'] = request.getHeader('Destination')
        # Parse request (if necessary)
        content_length = request.getHeader('Content-Length')
        if content_length is None or content_length == '0':
            return result
        else:
            doc = ElementTree.parse(request.content)
            root = doc.getroot()
            # Find behavior
            # TODO: Raise error if keepalive and omit are both present
            if root.find('{DAV:}keepalive') is not None:
                keepalive_elem = root.find('{DAV:}keepalive')
                if keepalive_elem.text == '*':
                    result['keepalive'] = '*'
                else:
                    result['keepalive'] = []
                    for child in keepalive_elem:
                        if child.tag == '{DAV:}href':
                            result['keepalive'].append(child.text)
            elif root.find('{DAV:}omit') is not None:
                result['omit'] = True
            return result

def create_server(reactor, engine, blob_server_list,
                  port=8080, service_name=None):
    log.debug("Creating WebDAV server...")
    root = ZFNResource(engine, blob_server_list)
    site = server.Site(root)
    log.info("Establishing WebDAV Bonjour service...")
    discovery.broadcast(reactor, '_webdav._tcp', port, service_name)
    log.debug("Acquiring port %i...", port)
    reactor.listenTCP(port, site)
