#!/usr/bin/env python
#
#   zfilenet/tests/test_index.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Testing module for the index"""

import unittest

from zfilenet.indexserver.index import (connect, create_session,
                                        IndexObject, File, Directory,
                                        get_root, resolve_path, move)

__author__ = 'Ross Light'
__date__ = 'July 20, 2008'
__all__ = ['DataTest']

class DataTest(unittest.TestCase):
    def setUp(self):
        self.engine = connect(':memory:')
        self.session = create_session(bind=self.engine)
        self.root = get_root(self.session)
    
    def tearDown(self):
        self.session.close()
    
    def test_clone(self):
        """File cloning should have the same information"""
        src_file = File("lumberjack.txt")
        clone = src_file.clone()
        assert clone.uuid != src_file.uuid, "UUIDs should not match"
        assert clone.name == src_file.name, "Names do not match"
    
    def test_directory_stat(self):
        """Directory stat test"""
        from zfilenet.util import to_uuid, format_datetime
        self.root.touch()
        self.root.touch_access()
        result = self.root.stat()
        assert result['type'] == 'directory', "Type is wrong"
        assert result['uuid'] == str(self.root.uuid), "UUID is wrong"
        assert result['created'] == format_datetime(self.root.created), \
            "Creation date is wrong"
        assert result['modified'] == format_datetime(self.root.modified), \
            "Modification date is wrong"
        assert result['accessed'] == format_datetime(self.root.accessed), \
            "Access date is wrong"
    
    def test_file_stat(self):
        """File stat test"""
        from zfilenet.util import to_uuid, format_datetime, data2hex
        new_file = File('blab')
        new_file.touch()
        new_file.touch_access()
        result = new_file.stat()
        assert result['type'] == 'file', "Type is wrong"
        assert result['uuid'] == str(new_file.uuid), "UUID is wrong"
        assert result['created'] == format_datetime(new_file.created), \
            "Creation date is wrong"
        assert result['modified'] == format_datetime(new_file.modified), \
            "Modification date is wrong"
        assert result['accessed'] == format_datetime(new_file.accessed), \
            "Access date is wrong"
        assert result['length'] == 0, "Length is wrong"
        assert result['sha1_hash'] is None, "Hash should not be present"
    
    def test_path(self):
        """File path test"""
        new_directory = Directory.make(self.session, '/spam')
        assert new_directory.get_path() == '/spam/', "Wrong directory path"
        new_file = File.make(self.session, '/spam/eggs.txt')
        assert new_file.get_path() == '/spam/eggs.txt', "Wrong file path"
    
    def test_resolve_path(self):
        """Path resolution test"""
        assert resolve_path(self.session, '/') is self.root, \
            "Root doesn't resolve"
        new_directory = Directory.make(self.session, '/spam')
        assert resolve_path(self.session, '/spam') is new_directory, \
            "New directory doesn't resolve"
        new_file = File.make(self.session, '/spam/eggs.txt')
        assert resolve_path(self.session, '/spam/eggs.txt') is new_file, \
            "New file doesn't resolve"
