#!/usr/bin/env python
#
#   zfilenet/tests/test_util.py
#   zfilenet
#
#   Copyright (C) 2008 Ross Light
#
#   This file is part of zfilenet.
#
#   zfilenet is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   zfilenet is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with zfilenet.  If not, see <http://www.gnu.org/licenses/>.
#

"""Testing module for the utilities module"""

import unittest

from zfilenet import util
from zfilenet.util import Path

__author__ = 'Ross Light'
__date__ = 'July 19, 2008'
__all__ = ['PathTest',
           'test_to_uuid',
           'test_data2hex',
           'test_hex2data',
           'test_parse_qname',
           'test_format_file_size',
           'test_format_datetime',]

class PathTest(unittest.TestCase):
    def test_creation(self):
        """Path creation test"""
        def _check_creation(path, components, absolute, directory):
            assert list(path.components) == list(components), \
                "Invalid components (got %r, should be %r)" % \
                (path.components, components)
            assert path.absolute == absolute, \
                "Wrong absoluteness (got %r, should be %r)" % \
                (path.absolute, absolute)
            assert path.directory == directory, \
                "Wrong directory-ness (got %r, should be %r)" % \
                (path.directory, directory)
        test_data = [
            [Path(), [], False, False],
            [Path(''), [], False, False],
            [Path('/'), [], True, False],
            [Path('//'), [], True, False],
            [Path('/spam'), ['spam'], True, False],
            [Path('eggs/'), ['eggs'], False, True],
            [Path('spam/eggs/bacon'), ['spam', 'eggs', 'bacon'], False, False],
            [Path('spam/../eggs'), ['eggs'], False, False],
            [Path('../bacon/eggs'), ['..', 'bacon', 'eggs'], False, False],
            [Path('/../sausage'), ['sausage'], True, False],
            [Path('ni/.'), ['ni'], False, False],
            [Path(['spam', 'eggs']), ['spam', 'eggs'], False, False],
            [Path(['spam', 'eggs'], True, True), ['spam', 'eggs'], True, True],
        ]
        for args in test_data:
            _check_creation(*args)
    
    def test_relative(self):
        """Relative paths resolution test"""
        path1 = Path('/home/python/')
        path2 = Path('spam/eggs')
        assert path1.relative_path(path2) == Path('/home/python/spam/eggs'), \
            "Simple resolution doesn't append"
        assert path2.relative_path('hello') == Path('spam/hello'), \
            "File resolution doesn't work"
        assert path2.relative_path(path1) == path1, \
            "Absolute paths don't pass thru"
    
    def test_convert(self):
        """Path conversions should change flags"""
        path_comps = ['spam', 'eggs']
        path = Path(path_comps)
        # Test absolute
        result = path.convert(absolute=True)
        assert result.absolute, "Did not change absolute"
        assert not result.directory, "Touched directory"
        # Test directory
        result = path.convert(directory=True)
        assert not result.absolute, "Touched absolute"
        assert result.directory, "Did not change directory"
        # Test both
        result = path.convert(absolute=True, directory=True)
        assert result.absolute, "Did not change absolute"
        assert result.directory, "Did not change directory"
    
    def test_string(self):
        """Path string test"""
        test_data = [
            [Path(), ''],
            [Path(''), ''],
            [Path('/'), '/'],
            [Path('//'), '/'],
            [Path('/spam'), '/spam'],
            [Path('eggs/'), 'eggs/'],
            [Path('spam/eggs/bacon'), 'spam/eggs/bacon'],
            [Path('spam/../eggs'), 'eggs'],
            [Path('../bacon/eggs'), '../bacon/eggs'],
            [Path('/../sausage'), '/sausage'],
            [Path('ni/.'), 'ni'],
            [Path('ni//blab///knights'), 'ni/blab/knights'],
        ]
        for (test, expected) in test_data:
            assert str(test) == expected, "Wrong path string"
    
    def test_add(self):
        """Path concatenation test"""
        assert Path('spam') + Path('eggs') == Path('spam/eggs'), \
            "Simple concatenation failed"
        assert Path('/spam') + Path('eggs') == Path('/spam/eggs'), \
            "Absolute concatenation failed"
        assert Path('spam') + Path('eggs/') == Path('spam/eggs/'), \
            "Directory concatenation failed"
        assert Path('spam') + 'eggs' == Path('spam/eggs'), \
            "String concatenation failed"
        assert 'knights' + Path('ni') == Path('knights/ni'), \
            "Reverse string concatenation failed"
    
    def test_length(self):
        """Path length should be the number of components"""
        assert len(Path('/')) == 0, "Wrong zero length"
        assert len(Path('bacon/eggs/spam')) == 3, "Wrong path length"
    
    def test_contains(self):
        """Paths should be able to test for containment"""
        assert 'eggs' in Path('bacon/eggs/spam'), "Component not found"
        assert 'knights' not in Path('bacon/eggs/spam'), "Component found"
    
    def test_comparison(self):
        """Path comparison test"""
        assert Path('bacon/eggs') == Path('bacon/eggs'), \
            "Paths should be equal"
        assert Path('bacon/eggs') != Path('eggs/bacon'), \
            "Paths should not be equal"
        assert Path('a/b/c') < Path('b/c'), "Path should be less"
        assert Path('a/b') < Path('a/b/c'), "Path should be less"
        assert Path('a/b') < Path('a/b/'), "Path should be less"
        assert Path('b/c') > Path('a/b/c'), "Path should be greater"
        assert Path('a/b/c') > Path('a/b'), "Path should be greater"
        assert Path('a/b/') > Path('a/b'), "Path should be greater"

def test_to_uuid():
    """to_uuid function test"""
    from uuid import UUID
    from zfilenet.util import to_uuid
    def _check_uuid(value):
        result = to_uuid(value)
        expected = UUID("01234567-89ab-cdef-0123-456789abcdef")
        assert result == expected, \
            "Wrong UUID created (got %r)" % (result)
    test_data = [
        ["01234567-89ab-cdef-0123-456789abcdef", "string"],
        ["0123456789abcdef0123456789abcdef", "jammed"],
        ["\x01\x23\x45\x67\x89\xab\xcd\xef\x01\x23\x45\x67\x89\xab\xcd\xef",
         "byte"],
        [0x0123456789abcdef0123456789abcdefL, "int"],
        [UUID("01234567-89ab-cdef-0123-456789abcdef"), "object"],
        [[0x01234567L, 0x89abL, 0xcdefL, 0x01L, 0x23L, 0x456789abcdefL],
         "field"],
    ]
    for (value, description) in test_data:
        _check_uuid.description = "to_uuid %s test" % description
        yield _check_uuid, value

def test_data2hex():
    """data2hex function test"""
    from zfilenet.util import data2hex
    assert data2hex('hello\x00\xff\x7f') == '68656c6c6f00ff7f', \
        "Wrong data2hex conversion"

def test_hex2data():
    """hex2data function test"""
    from zfilenet.util import hex2data
    assert hex2data('68656c6c6f00ff7f') == 'hello\x00\xff\x7f', \
        "Wrong hex2data conversion"

def test_parse_qname():
    """parse_qname function test"""
    from zfilenet.util import parse_qname
    assert parse_qname('{spam}eggs') == ('spam', 'eggs'), \
        "Qualified name mismatch"
    assert parse_qname('bacon') == (None, 'bacon'), "Plain name mismatch"
    assert parse_qname('{}ni') == ('', 'ni'), "Empty namespace mismatch"

def test_format_file_size():
    """File size formatting test"""
    from zfilenet.util import format_file_size as format
    def _check_size(value, expected):
        assert format(value) == expected, \
            "Wrong formatting (expected %r)" % expected
    test_data = [
        [0, "0 bytes", "zero"],
        [42, "42 bytes", "bytes"],
        [1.2 * 1024, "1.2 KiB", "kibibyte"],
        [97.8 * (1024 ** 2), "97.8 MiB", "mibibyte"],
        [42.3 * (1024 ** 3), "42.3 GiB", "gibibyte"],
        [42.3 * (1024 ** 4), "42.3 TiB", "tibibyte"],
        [3.14 * (1024 ** 5), "3.1 PiB", "pitibyte"],
    ]
    for (size, expected, name) in test_data:
        _check_size.description = "File size formatting %s test" % name
        yield _check_size, int(size), expected

def test_format_datetime():
    """Datetime formatting test"""
    from datetime import datetime
    from zfilenet.util import format_datetime
    dt = datetime(2008, 4, 27, 14, 32, 54, 314159)
    assert format_datetime(dt) == '2008-04-27T14:32:54Z', "Bad formatting"
